from trame_server.utils.asynchronous import (
    create_task,
    decorate_task,
    create_state_queue_monitor_task,
    StateQueue,
    task,
)

__all__ = [
    "create_task",
    "decorate_task",
    "create_state_queue_monitor_task",
    "StateQueue",
    "task",
]



r"""
From a list of trame modules name, gather and generate the required static
content that needs to be served for a trame application to work.
"""
import argparse
import importlib
import shutil
from pathlib import Path


class StaticContentGenerator:
    def __init__(self):
        self.www = None
        self.serve = {}

    def enable_all(self):
        import pkgutil

        root = importlib.import_module("trame.modules")
        self.enable_modules(*[m.name for m in pkgutil.iter_modules(root.__path__)])

    def enable_modules(self, *names):
        for module_name in names:
            module = None
            try:
                module = importlib.import_module(f"trame.modules.{module_name}")
            except ModuleNotFoundError:
                try:
                    print("module_name:", module_name)
                    module = importlib.import_module(module_name)
                except ModuleNotFoundError:
                    print(f" - Error: Skipping module {module_name}")

            if module is None:
                print(f"Skipping module: {module_name}")
                continue

            if "serve" in module.__dict__:
                self.serve.update(module.serve)

            if "www" in module.__dict__:
                self.www = module.www

    def write(self, output_directory=None):
        if output_directory is None:
            output_directory = Path.cwd()

        if self.www:
            shutil.copytree(self.www, output_directory, dirs_exist_ok=True)

        for sub_path, src_dir in self.serve.items():
            dst_dir = Path(output_directory) / sub_path
            shutil.copytree(src_dir, dst_dir, dirs_exist_ok=True)


def main():
    parser = argparse.ArgumentParser(
        description="Static Web Client generator for trame applications"
    )

    parser.add_argument(
        "--output",
        help="Directory to fill with trame client code",
    )

    args, module_names = parser.parse_known_args()
    generator = StaticContentGenerator()
    if len(module_names):
        generator.enable_modules(*module_names)
    else:
        generator.enable_all()
    generator.write(args.output)


if __name__ == "__main__":
    main()



class ClientFile:
    def __init__(self, file_state_variable=None):
        self._name = None
        self._size = None
        self._time = None
        self._mime_type = None
        self._content = None
        if file_state_variable is not None:
            self._name = file_state_variable.get("name")
            self._size = file_state_variable.get("size")
            self._time = file_state_variable.get("lastModified")
            self._mime_type = file_state_variable.get("type")
            self._content = file_state_variable.get("content")
            if isinstance(self._content, list):
                self._content = b"".join(self._content)

    @property
    def is_empty(self):
        """Return true if the file is empty"""
        return self._content is None

    @property
    def name(self):
        """File name"""
        return self._name

    @property
    def size(self):
        """File size in Bytes"""
        return self._size

    @property
    def modified_time(self):
        """Modified time"""
        return self._time

    @property
    def mime_type(self):
        """Mime type of the file"""
        return self._mime_type

    @property
    def content(self):
        """Bytes content of the file"""
        return self._content

    @property
    def info(self):
        return f"File: {self.name} of size {self.size} and type {self.mime_type}"



def clear_triggers(server):
    """
    Helper function to remove all triggers.

    :param server: server on which we want to clear the triggers
    :type server: trame_server.core.Server
    """
    names = list(server._triggers.keys())
    for name in names:
        fn = server._triggers.pop(name)
        server._triggers_fn2name.pop(fn)
        print(f"unregister trigger {name}")


def clear_change_listeners(server):
    """
    Helper function to remove all state.change listeners.

    :param server: server on which we want to clear the state.change listeners
    :type server: trame_server.core.Server
    """
    server._change_callbacks.clear()


def remove_change_listeners(server, *names):
    """
    Helper function to remove any listeners for a given set
    of state variable names.

    :param server: server on which we want to clear the state.change listeners
    :type server: trame_server.core.Server

    :param *names: State variable names
    :type *names: str
    """
    for name in names:
        if name in server._change_callbacks:
            server._change_callbacks.pop(name)


def reload(*reload_list):
    """
    Helper function use to reload python modules that were passed as
    arguments.

    :param *reload_list: positional arguments of the modules to reload when the
                        reload button is pressed.
    :type *reload_list: python modules
    """
    for m in reload_list:
        m.__loader__.exec_module(m)



import asyncio
from trame.app import get_server
from IPython import display

__all__ = [
    "show",
    "display_iframe",
    "run",
]


def show(_server, ui=None, **kwargs):
    """
    Helper function to show a server ui element into the cell.

    :param _server: the server on which the UI is defined
    :type _server: trame_server.core.Server

    :param ui: the name of the ui section to display. (Default: 'main')
    :type ui: str

    :param **kwargs: any keyword arguments are pass to the Jupyter IFrame.
        Additionally `protocol=` and `host=` can be use to override the iframe src url.
    """
    if isinstance(_server, str):
        _server = get_server(_server)

    def on_ready(**_):
        params = f"?ui={ui}" if ui else ""
        src = f"{kwargs.get('protocol', 'http')}://{kwargs.get('host', 'localhost')}:{_server.port}/index.html{params}"
        loop = asyncio.get_event_loop()
        loop.call_later(0.1, lambda: display_iframe(src, **kwargs))
        _server.controller.on_server_ready.discard(on_ready)

    if _server._running_stage == 0:
        _server.controller.on_server_ready.add(on_ready)
        _server.start(
            exec_mode="task",
            port=0,
            open_browser=False,
            show_connection_info=False,
            disable_logging=True,
            timeout=0,
        )
    elif _server._running_stage == 1:
        _server.controller.on_server_ready.add(on_ready)
    elif _server._running_stage == 2:
        on_ready()


def display_iframe(src, **kwargs):
    """
    Convenience method to display an iframe for the given url source

    :param src: url to display
    :type src: str

    :param **kwargs: any keyword arguments are pass to the Jupyter IFrame.
    """

    # Set some defaults. The kwargs can override these.
    # width and height are both required.
    iframe_kwargs = {
        "width": "100%",
        "height": 600,
        **kwargs,
    }
    iframe = display.IFrame(src=src, **iframe_kwargs)
    return display.display(iframe)


def run(name, **kwargs):
    """Run and display a Jupyter server proxy process with the given name

    Note that the proxy process must be registered with Jupyter by setting
    the `jupyter_serverproxy_servers` entrypoint in its setup.py or setup.cfg
    file.
    """
    src = f"/{name}"
    return display_iframe(src, **kwargs)



import mimetypes


MIMETYPE_OVERRIDES = {
    # On Windows, mimetypes pulls this from the registry, which is
    # wrong. Override it.
    "application/javascript": ".js",
}


def decorate_mimetypes_init():
    """
    Decorate the mimetypes.init() method with our function that
    afterwards adds any mimetype overrides that were saved.
    """

    original_init = mimetypes.init

    def new_init(*args, **kwargs):
        original_init(*args, **kwargs)
        for key, value in MIMETYPE_OVERRIDES.items():
            mimetypes.add_type(key, value)

    mimetypes.init = new_init


def add_mimetype_override(type, ext):
    """
    Add a mimetype both now and when mimetypes.init() is called.

    :param type: mimetype
    :type type: str
    :param ext: file extension for which the mimetype applies
    :type ext: str
    """
    # Add it to the overrides that are performed if init() is called
    MIMETYPE_OVERRIDES[type] = ext

    # Also override it right now
    mimetypes.add_type(type, ext)


def to_mime(file_path):
    """
    Return the mime type from a given path

    :param file_path: Path to analyse
    :type file_path: str

    :return: Mime type
    :rtype: str
    """
    return mimetypes.guess_type(file_path)[0]


# Decorate the mimetypes.init() method
decorate_mimetypes_init()



import base64
from pathlib import Path

from trame.app.mimetypes import to_mime


def to_txt(full_path):
    """
    Return the text content of the file path

    :param file_path: Path to the file to read
    :type file_path: str

    :return: File content
    :rtype: str
    """
    with open(full_path) as txt_file:
        return str(txt_file.read())


def to_base64(full_path):
    """
    Return the base64 content of the file path

    :param file_path: Path to the file to read
    :type file_path: str

    :return: File content encoded in base64
    :rtype: str
    """
    with open(full_path, "rb") as bin_file:
        return base64.b64encode(bin_file.read()).decode("ascii")


def to_url(full_path):
    """
    Return the base64 encoded URL of the file path

    :param file_path: Path to the file to read
    :type file_path: str

    :return: Inlined bas64 encoded url (data:{mime};base64,{content})
    :rtype: str
    """
    encoded = to_base64(full_path)
    mime = to_mime(full_path)
    return f"data:{mime};base64,{encoded}"


class LocalFileManager:
    """LocalFileManager provide convenient methods for handling local files"""

    def __init__(self, base_path):
        """
        Provide the base path on which relative path should be based on.

        :param base_path: A file or directory path
        :type base_path: str
        """
        _base = Path(base_path)

        # Ensure directory
        if _base.is_file():
            _base = _base.parent

        self._root = Path(str(_base.resolve().absolute()))
        self._assests = {}

    def __getitem__(self, name):
        return self._assests.get(name)

    def __getattr__(self, name):
        return self._assests.get(name)

    def _to_path(self, file_path):
        _input_file = Path(file_path)
        if _input_file.is_absolute():
            return str(_input_file.resolve().absolute())

        return str(self._root.joinpath(file_path).resolve().absolute())

    def base64(self, key, file_path=None):
        """
        Store a base64 file content under the provided key name

        :param key: name for that content which can then be accessed
                    by the [] or . notation
        :type key: str

        :param file_path: A file or directory path
        :type file_path: str

        :return: Return the stored content
        :rtype: str
        """
        if file_path is None:
            file_path, key = key, file_path

        data = to_base64(self._to_path(file_path))

        if key is not None:
            self._assests[key] = data

        return data

    def url(self, key, file_path):
        """
        Store a url encoded file content under the provided key name

        :param key: name for that content which can then be accessed
                    by the [] or . notation
        :type key: str

        :param file_path: A file or directory path
        :type file_path: str

        :return: Return the stored content
        :rtype: str
        """
        if file_path is None:
            file_path, key = key, file_path

        data = to_url(self._to_path(file_path))

        if key is not None:
            self._assests[key] = data

        return data

    def txt(self, key, file_path):
        """
        Store a file content (text) under the provided key name

        :param key: name for that content which can then be accessed
                    by the [] or . notation
        :type key: str

        :param file_path: A file or directory path
        :type file_path: str

        :return: Return the stored content
        :rtype: str
        """
        if file_path is None:
            file_path, key = key, file_path

        data = to_txt(self._to_path(file_path))

        if key is not None:
            self._assests[key] = data

        return data

    @property
    def assets(self):
        """Return the full set of assets as a dict"""
        return self._assests

    def get_assets(self, *keys):
        """Return a filtered out dict using the provided set of keys"""
        if len(keys) == 0:
            return self.assets

        _assets = {}
        for key in keys:
            _assets[key] = self._assests.get(key)

        return _assets


__all__ = [
    "LocalFileManager",
    "to_mime",
    "to_txt",
    "to_base64",
    "to_url",
]
